package enverstest.model;

import static org.junit.Assert.*;

import java.util.List;

import javax.persistence.EntityManager;

import org.hibernate.envers.AuditReader;
import org.hibernate.envers.AuditReaderFactory;
import org.hibernate.envers.query.AuditEntity;
import org.hibernate.envers.query.AuditQuery;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;

import enverstest.util.JpaUtil;

public class CategoryTest
{
	private EntityManager em;
	
	// set the following variable to true to demonstrate the problem
	// in case a mapping with "composite-element" tag is used.
	// Note: to actually demonstrate the problem, you have to modify
	// Category.hbm.xml to use the "composite-element" mapping
	private boolean showEnversQueryExceptionError = false;

	// set the following variable to true to demonstrate the problem
	// with nullable properties in the mapping with CompositeUserType
	private boolean showEnversNullablePropertyError = false;
	
	@Before
	public void setUp() throws Exception 
	{
		em = JpaUtil.getEntityManager("enverstest");
	}

	@After
	public void tearDown() throws Exception 
	{
		JpaUtil.shutdown();
	}

	@Test
	public void testEnvers() 
	{
		long startPIT = System.currentTimeMillis();
		
		if (showEnversQueryExceptionError) {
			System.out.println("\nDemonstrating the QueryException problem.");
			System.out.println("(Make sure Category.hbm.xml uses the \"custom-element mapping\"!)");
		}
		
		if (showEnversNullablePropertyError) {
			System.out.println("\nDemonstrating the nullable property problem.");
			System.out.println("(Make sure Category.hbm.xml uses the \"element with custom type mapping\"!)");
		}
		
		if (!showEnversQueryExceptionError && !showEnversNullablePropertyError) {
			System.out.println("\nAll problem demonstrations are disabled");
			System.out.println("The test should run successfully!");
			
		}
		System.out.println("\n===== Fill Database =====");

		JpaUtil.beginTransaction();
		
		Item item = new Item("The Item");
		
		assertNull(item.getId());
		
		em.persist(item);
		
		assertNotNull(item.getId());
		
		Category category = new Category("The Category");
		
		category.setDescription("The description");
		
		// create value object with one property set to null if demonstration
		// of the "nullable problem" is requested
		Long numberValue = (showEnversNullablePropertyError) ? null : 4711L;
		
		category.setValue(item, new Value("The Value", numberValue));
			
		category.setText(item, "The Text");
		
		assertNull(category.getId());
		
		em.persist(category);
		
		assertNotNull(category.getId());
		
		JpaUtil.commitTransaction();
		
		System.out.println("\n===== Database filled =====");

		em.clear();
		
		System.out.println("\n===== Check current entities =====");
		
		JpaUtil.beginTransaction();
		
		Category loadedCategory = em.find(Category.class, category.getId());
		
		assertNotNull(loadedCategory);
		assertEquals("The Category", loadedCategory.getName());
		assertEquals("The description", loadedCategory.getDescription());
		
		Value theValue = loadedCategory.getValue(item);
		assertNotNull(theValue);
		assertEquals("The Value", theValue.getText());
		assertEquals(numberValue, theValue.getNumber());
		
		assertEquals("The Text", loadedCategory.getText(item));
		
		JpaUtil.commitTransaction();
		
		System.out.println("\n===== Load historic data =====");
		
		JpaUtil.beginTransaction();
		
		// first we retrieve the first revision of the Category entity
		AuditReader reader = AuditReaderFactory.get(em);
		AuditQuery query = reader.createQuery().forRevisionsOfEntity(Category.class, false, true)
		                                       .addOrder(AuditEntity.revisionProperty("timestamp").asc())
		                                       .add(AuditEntity.id().eq(category.getId()));
		
		@SuppressWarnings("unchecked")
		List<Object[]> history = (List<Object[]>)query.getResultList();
		
		assertNotNull(history);
		assertEquals(1, history.size());
		
		Revision first = (Revision)history.get(0)[1];
		
		assertNotNull(first);
		assertTrue(startPIT < first.getTimestamp().longValue());
		
		// Now read historic data of the category entity at the first revision
		// This will of course be the same as the current entity, but it is enough
		// to show the problem
		query = reader.createQuery().forEntitiesAtRevision(Category.class, first.getId())
		                            .add(AuditEntity.property("id").eq(category.getId()))
		                            .setMaxResults(1);
		
		Category historicCategory = (Category)query.getSingleResult();
		
		assertNotNull(historicCategory);
		
		System.out.println("\n===== Check historic data =====");
		
		assertEquals(category.getId(), historicCategory.getId());
		assertEquals("The Category", historicCategory.getName());
		assertEquals("The description", historicCategory.getDescription());

		// this assertion holds and does not produce an exception,
		// even though there is also a map association involved!
		// But this map does not use a composite-element tag!
		assertEquals("The Text", historicCategory.getText(item));

		
		/*
		 * If you set showEnversError to "true", the following code will produce an
		 * 
		 * org.hibernate.QueryException: could not resolve property: originalId.rev_type of: enverstest.model.Item_r
		 * 
		 * if the property "categoryItem" is mapped with the "composite-element" tag.
		 * If the property is mapped with the "element" tag using its own CompositeUserType 
		 * implementation no exception occurs!
		 * 
		 * The query produced by Envers and rejected by Hibernate is:
		 * 
		 * select new list(ee__, e__) 
		 *   from category_item_r ee__, 
		 *        enverstest.model.Item_r e__ 
		 *  where ee__.originalId.mapkey_id = e__.originalId.id 
		 *    and ee__.originalId.Category_id = :Category_id 
		 *    and e__.originalId.rev_begin.id <= :revision 
		 *    and ee__.originalId.rev_begin.id <= :revision 
		 *    and ee__.originalId.rev_type != :delrevisiontype 
		 *    and e__.originalId.rev_type != :delrevisiontype 
		 *    and (e__.rev_end.id > :revision or e__.rev_end is null) 
		 *    and (ee__.rev_end.id > :revision or ee__.rev_end is null)
		 *    
		 * If you set showEnversNullablePropertyError to "true", the following code will produce an
		 * 
		 * javax.persistence.EntityNotFoundException: Unable to find category_item_r with id {mapkey_id=15, rev_begin=enverstest.model.Revision@6f3e19b3, Category_id=15, element=enverstest.model.Value@6ed16657}
		 * 
		 * if the property "categoryItem" is mapped with the composite user type ValueType
		 * (and the "value" property is set to a null value).
		 * 
		 * In that case, Envers/Hibernate creates the following SQL query:
		 * 
		 * select category_i0_.rev_begin as rev_begi1_2_0_, 
		 *        category_i0_.category_id as category2_2_0_, 
		 *        category_i0_.number as number3_2_0_, 
		 *        category_i0_.text as text4_2_0_, 
		 *        category_i0_.item_id as item_id5_2_0_, 
		 *        category_i0_.rev_type as rev_type6_2_0_, 
		 *        category_i0_.rev_end as rev_end7_2_0_ 
		 *   from category_item_r category_i0_ 
		 *  where category_i0_.rev_begin=? 
		 *    and category_i0_.category_id=? 
		 *    and category_i0_.number=? 
		 *    and category_i0_.text=? 
		 *    and category_i0_.item_id=?
		 *    
		 * The problem here is that the where clause checks for a null value with
		 * "category_i0_.number=null", which leads to an empty resultset.
		 * The correct where clause would be: "category_i0_.number is null"
		 */
		if (showEnversQueryExceptionError || showEnversNullablePropertyError) {
			Value historicValue = historicCategory.getValue(item);
			assertNotNull(historicValue);
			assertEquals("The Value", historicValue.getText());
			assertEquals(numberValue, historicValue.getNumber());
		}
		
		System.out.println("\n===== SUCCESS - all data is valid and no exception occurred! =====\n");
		
		JpaUtil.commitTransaction();
	}
}
